
import org.eclipse.swt.all;

import java.lang.all;

void createScrollBasePoint(Scrollable pane, int hSpan = 1, int vSpan = 1) {
	auto d = pane.getDisplay();
	auto shell = new Shell(d, SWT.NO_TRIM|SWT.ON_TOP);
	shell.setBackground(d.getSystemColor(SWT.COLOR_WHITE));
	auto region = new Region(d);
	static immutable R = 18;
	static immutable R2 = R * 2;
	auto image = new Image(d, R2, R2);
	scope (exit) image.dispose();
	auto gc = new GC(image);
	scope (exit) gc.dispose();
	gc.setBackground(d.getSystemColor(SWT.COLOR_BLACK));
	gc.fillOval(0, 0, R2, R2);
	auto imgData = image.getImageData();
	foreach (x; 0..imgData.width) {
		foreach (y; 0..imgData.height) {
			if (!imgData.getPixel(x, y)) {
				region.add(x, y, 1, 1);
			}
		}
	}
	shell.setRegion(region);
	shell.setAlpha(192);
	auto triY1 = 4;
	auto triY2 = R - R / 4 - 4;
	auto triX1 = R;
	auto triX2 = R + R / 3;
	auto triN = [triX1, triY1, triX2, triY2, R2 - triX2, triY2];
	auto triE = [triY1, triX1, triY2, triX2, triY2, R2 - triX2];
	auto triS = [triX1, R2 - triY1, triX2, R2 - triY2, R2 - triX2, R2 - triY2];
	auto triW = [R2 - triY1, triX1, R2 - triY2, triX2, R2 - triY2, R2 - triX2];
	shell.addPaintListener(new class PaintListener {
		override void paintControl(PaintEvent e) {
			e.gc.setForeground(d.getSystemColor(SWT.COLOR_BLACK));
			e.gc.setBackground(d.getSystemColor(SWT.COLOR_BLACK));
			e.gc.setLineWidth(2);
			e.gc.setAntialias(SWT.ON);
			e.gc.drawOval(2, 2, R2 - 4, R2 - 4);
			e.gc.fillOval(R - 4, R - 4, 8, 8);
			auto vbar = pane.getVerticalBar();
			auto hbar = pane.getHorizontalBar();
			if (vbar && vbar.isVisible()) {
				e.gc.fillPolygon(triN);
				e.gc.fillPolygon(triS);
			}
			if (hbar && hbar.isVisible()) {
				e.gc.fillPolygon(triW);
				e.gc.fillPolygon(triE);
			}
		}
	});
	bool closed = true;
	auto move = new class Runnable {
		override void run() {
			if (closed) return;
			if (shell.isDisposed()) return;
			if (pane.isDisposed()) return;

			if (!cast(Table)pane) pane.setRedraw(false);
			scope (exit) {
				if (!cast(Table)pane) pane.setRedraw(true);
			}

			auto vbar = pane.getVerticalBar();
			auto hbar = pane.getHorizontalBar();
			auto curPos = d.getCursorLocation();
			auto b = shell.getBounds();
			auto sPosX = b.x + b.width / 2;
			auto sPosY = b.y + b.height / 2;
			int hMove = 0, vMove = 0;
			if (R2 < sPosX - curPos.x) {
				hMove = (curPos.x - sPosX) / R2;
			}
			if (-R2 > sPosX - curPos.x) {
				hMove = (curPos.x - sPosX) / R2;
			}
			if (R2 < sPosY - curPos.y) {
				vMove = (curPos.y - sPosY) / R2;
			}
			if (-R2 > sPosY - curPos.y) {
				vMove = (curPos.y - sPosY) / R2;
			}
			import org.eclipse.swt.internal.win32.OS;
			import org.eclipse.swt.internal.win32.WINTYPES;
			if (hMove != 0) {
				if (cast(Table)pane) {
					OS.SendMessage(pane.handle, OS.LVM_SCROLL, hMove * hSpan, 0);
				} else {
					int pos = hbar.getSelection() + hMove * hSpan;
					hbar.setSelection(pos);
				}
			}
			if (vMove != 0) {
				if (auto table = cast(Table)pane) {
					int index = table.getTopIndex();
					(cast(Table)pane).setTopIndex(index + vMove * vSpan);
				} else {
					int pos = vbar.getSelection() + vMove * vSpan;
					vbar.setSelection(pos);
				}
			}
		}
	};
	void close() {
		closed = true;
		shell.setVisible(false);
	}
	pane.addDisposeListener(new class DisposeListener {
		override void widgetDisposed(DisposeEvent e) {
			close();
			synchronized (pane) {
				d = null;
			}
			shell.dispose();
			region.dispose();
		}
	});
	pane.addFocusListener(new class FocusAdapter {
		override void focusLost(FocusEvent e) { close(); }
	});
	pane.addKeyListener(new class KeyAdapter {
		override void keyPressed(KeyEvent e) { close(); }
	});
	pane.addMouseListener(new class MouseAdapter {
		override void mouseDown(MouseEvent e) {
			if (!closed) {
				close();
			} else if (e.button == 2) {
				auto curPos = d.getCursorLocation();
				shell.setBounds(curPos.x - R, curPos.y - R, R2, R2);
				shell.setVisible(true);
				closed = false;
				auto thr = new core.thread.Thread({
					while (!closed) {
						import std.datetime;
						core.thread.Thread.sleep(dur!"msecs"(100));
						synchronized (pane) {
							if (d) d.asyncExec(move);
						}
					}
				});
				thr.start();
			}
		}
	});
}

void main() {
	import std.string;
	import std.array;
	import std.random;

	auto display = new Display();
	auto shell = new Shell(display, SWT.SHELL_TRIM|SWT.DOUBLE_BUFFERED);
	shell.setLayout(new FillLayout);

	void createWidget(T)() {
		auto w = new T(shell, SWT.NONE);
		foreach (i; 0..100) {
			auto item = mixin(`new ` ~ T.stringof ~ `Item(w, SWT.NONE)`);
			item.setText("Lo%sng Item %s".format("o".replicate(uniform(0, 35)), i));
		}
		createScrollBasePoint(w, w.getItem(0).getBounds().height, 1);
	}
	createWidget!Table();
	createWidget!Tree();

	shell.setSize(600, 300);
	shell.open();

	while (!shell.isDisposed()) {
		if (!display.readAndDispatch()) {
			display.sleep();
		}
	}
	display.dispose();
}
